from .cli import trigger_cli, register_cli, client_cli, server_cli
from .fetch import get_keys
from .plugins import PLUGINS, register_publish, register_response, register_trigger, load_plugins

from . import get_api

from .version import __version__



import configparser
from pathlib import Path

import zmq

from .defaults import (
    AUTHORISED_KEYS_FOLDER,
    DEFAULT_FOLDER,
    KNOWN_HOSTS_FOLDER,
    PLUGINS_FOLDER,
    DEFAULT_SERVER_CFG,
    DEFAULT_LOGGING,
    DEFAULT_TIMEOUTS,
    MAX_MESSAGE_SIZE
)


def gen_paths(cfg_path, key_path):
    ''' Generates the configuration paths based on the user's input. If no
        input, then default locations are used, otherwise the config paths are
        set relative to the inputted path

    Parameters
    ----------
    cfg_path : Posix path
        The path to the folder containing the instrument's config file
    key_path : Posix path
        The path to the folder containing the instrument's keypair

    Returns
    -------
    path_conf : Posix path
        The path to the folder containing the instrument's config file
    path_auth : Posix path
        The path to the folder containing the authorised public keys
    path_host : Posix path
        The path to the folder containing the available servers' public keys
    path_pair : Posix path
        The path to the folder containing the instrument's keypair
    path_plug : Posix path
        The path to the folder containing the plugins
    '''
    if cfg_path is None:
        path_conf = DEFAULT_FOLDER
        path_auth = AUTHORISED_KEYS_FOLDER
        path_host = KNOWN_HOSTS_FOLDER
        path_plug = PLUGINS_FOLDER
    else:
        path_conf = cfg_path
        path_auth = path_conf / "authorised_keys"
        path_host = path_conf / "known_hosts"
        path_plug = path_conf / "plugins"

    if key_path is None:
        path_pair = path_conf / "keypairs"
    else:
        path_pair = key_path

    if not path_conf.is_dir():
        path_conf.mkdir()
    if not path_auth.is_dir():
        path_auth.mkdir()
    if not path_host.is_dir():
        path_host.mkdir()
    if not path_pair.is_dir():
        path_pair.mkdir()
    if not path_plug.is_dir():
        path_plug.mkdir()

    return path_conf, path_auth, path_host, path_pair, path_plug


def gen_keys(name, cfg_path, key_path):
    ''' Generates the instrument's public-private keypair and places them in
        the keypairs folder

    Parameters
    ----------
    name : string
        The name of the server
    cfg_path : Posix path
        The path to the folder containing the instrument's config file
    key_path : Posix path
        The path to the folder containing the instrument's keypair
    '''
    file_pub, file_prv = zmq.auth.create_certificates(cfg_path, name)
    key_pub, key_prv = zmq.auth.load_certificate(file_prv)

    Path.rename(
        cfg_path / f"{name}.key",
        key_path / f"{name}.key"
    )
    Path.rename(
        cfg_path / f"{name}.key_secret",
        key_path / f"{name}-curve.key_secret"
    )


def gen_server_cfg(path_conf, path_pub, path_prv, path_plug):
    ''' Generates a configuration file for the server, containing the server's
        TCP/IP details (for both REP and PUB) socket types, as well as the
        locations of the folders holding the authorised public keys, private
        keys and plugins

    Parameters
    ----------
    path_conf : Posix path
        The path to the folder containing the server's config file
    path_pub : Posix path
        The path to the folder containing the authorised public keys
    path_prv : Posix path
        The path to the folder containing the server's private key
    path_plug : Posix path
        The path to the folder containing the plugins
    '''
    config = configparser.ConfigParser()

    config['Security'] = {
        'public-keys': path_pub,
        'private-keys': path_prv
    }
    config['Plugins'] = {
        'plugins': path_plug
    }

    for section in DEFAULT_LOGGING:
        config.add_section(section)
        for key in DEFAULT_LOGGING[section]:
            config[section][key] = DEFAULT_LOGGING[section][key]

    for section in DEFAULT_SERVER_CFG:
        config.add_section(section)
        for key in DEFAULT_SERVER_CFG[section]:
            config[section][key] = DEFAULT_SERVER_CFG[section][key]

    config["Messages"] = {
        'max-size': MAX_MESSAGE_SIZE
    }

    filename = path_conf / "server.cfg"
    with open(filename, 'w') as configfile:
        config.write(configfile)


def gen_client_cfg(path_conf, path_pub, path_prv):
    ''' Generates a configuration file for the client, containing the
        locations of the folders holding the public keys of the known hosts,
        private keys and plugins

    Parameters
    ----------
    path_conf : Posix path
        The path to the folder containing the client's config file
    path_pub : Posix path
        The path to the folder containing the known hosts' public keys
    path_prv : Posix path
        The path to the folder containing the client's private key
    '''
    config = configparser.ConfigParser()

    config['Security'] = {
        'public-keys': path_pub,
        'private-keys': path_prv
    }

    for section in DEFAULT_LOGGING:
        config.add_section(section)
        for key in DEFAULT_LOGGING[section]:
            config[section][key] = DEFAULT_LOGGING[section][key]

    for section in DEFAULT_TIMEOUTS:
        config.add_section(section)
        for key in DEFAULT_TIMEOUTS[section]:
            config[section][key] = DEFAULT_TIMEOUTS[section][key]

    filename = path_conf / "hosts.cfg"
    with open(filename, 'w') as configfile:
        config.write(configfile)



import configparser
from datetime import datetime
import logging
from pathlib import Path
import sys
import warnings

import zmq

from .defaults import (
    Client,
    Server,
    Paths,
    _CFG_PATHS_SERVER,
    _CFG_PATHS_CLIENT,
    DEFAULT_FOLDER,
    DEFAULT_TIMEOUTS
)
from .plugins import PLUGINS, load_plugins

logger = logging.getLogger(__name__)


def find_config(args):
    ''' Uses the input arguments to extract the folder containing the
        configuration file

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system

    Returns
    -------
    folder : Posix path
        The path to the folder containing the configuration file
    '''
    if args.cfgpath is None:
        folder = DEFAULT_FOLDER
        logger.debug(f"Path to config folder: {None}")
    else:
        folder = Path(args.cfgpath)
        logger.debug(f"Path to config folder: {folder}")

    return folder


def load_config(folder, container):
    ''' Establishes the path to the configuration file and reads its contents

    Parameters
    ----------
    folder : Posix path
        The path to the folder containing the configuration file
    container : string
        Whether the user is running a server or a client

    Returns
    -------
    config : ConfigParser
        The set of configs found in the configuration file
    '''
    if container == "server":
        config_file = folder / "server.cfg"
        defaults = _CFG_PATHS_SERVER
    elif container == "client":
        config_file = folder / "hosts.cfg"
        defaults = _CFG_PATHS_CLIENT

    config = configparser.ConfigParser()

    if config_file is not None:
        if not isinstance(config_file, Path):
            config_file = Path(config_file)
        config.read([config_file])

    for section, key in defaults:
        _path = Path(config.get(section, key)).resolve()
        config.set(section, key, str(_path))
        if not _path.exists():
            warnings.warn(f"configured path '{_path}' does not exist")

    return config


def setup_logging(args, config):
    ''' Uses the input arguments and configs to set up logging

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system
    config : ConfigParser
        The set of configs found in the client's configuration file
    '''
    lib_logger = logging.getLogger("rain")

    cfg_level = config.get("Logging", "level", fallback="INFO")
    cfg_file = config.get("Logging", "filepath", fallback=None)
    if cfg_file == "None":
        cfg_file = None
    cfg_filelevel = config.get("Logging", "file-level", fallback="INFO")
    cfg_print = config.getboolean("Logging", "print", fallback=False)
    cfg_printlevel = config.get("Logging", "print-level", fallback="INFO")

    if args.loglevel is not None:
        loglevel = args.loglevel
    else:
        loglevel = cfg_level

    if args.logfile is not None:
        logfile = args.logfile.resolve()
    else:
        logfile = cfg_file

    if args.filelevel is not None:
        filelevel = args.filelevel
    else:
        filelevel = cfg_filelevel

    logprint = cfg_print or args.logprint

    if args.printlevel is not None:
        printlevel = args.printlevel
    else:
        printlevel = cfg_printlevel

    # TODO 65: Make logging format configurable
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s - %(message)s')

    if logfile is not None:
        handler = logging.FileHandler(str(logfile))
        handler.setFormatter(formatter)
        handler.setLevel(filelevel)
        lib_logger.addHandler(handler)

    if logprint:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        handler.setLevel(printlevel)
        lib_logger.addHandler(handler)

    lib_logger.setLevel(loglevel)
    lib_logger.debug("Logging setup complete")


def find_paths(config, container):
    ''' Uses the configs from the configuration file to extract the paths to
        the public keys, keypair and (for a server) plugins

    Parameters
    ----------
    config : ConfigParser
        The set of configs found in the client's configuration file
    container : string
        Whether the user is running a server or a client

    Returns
    -------
    paths : rain.defaults.Paths
        An object containing the paths to the folders holding the senders'
        public keys, the user's private keypair and the server's plugins folder
        (if the user is a server)
    '''
    path_pub = Path(config.get("Security", "public-keys"))
    path_prv = Path(config.get("Security", "private-keys"))
    if container == "server":
        path_plug = Path(config.get("Plugins", "plugins"))
    else:
        path_plug = None

    paths = Paths(path_pub, path_prv, path_plug)

    return paths


def setup_server_object(args, config):
    ''' Reads the input arguments and the loaded configs to fill in a Server
        object containing relevant connection information

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system
    config : ConfigParser
        The set of configs found in the configuration file

    Returns
    -------
    server : rain.defaults.Server
        Contains information regarding the connection established by the server
    '''
    if "Allowed" in config:
        allowed = [config.get("Allowed", option) for option in config["Allowed"]]

    server = Server(args.host, allowed)

    if args.host == "rep":
        server.publ_host = config.get("Response", "hostname")
        server.publ_port = config.get("Response", "port")

    elif args.host == "pub":
        server.publ_host = config.get("Publish", "hostname")
        server.publ_port = config.get("Publish", "port")
        server.trig_host = config.get("Trigger", "hostname")
        server.trig_port = config.get("Trigger", "port")
        server.enable_auth = args.auth

    if "Messages" in config:
        server.max_msg_size = config.getint("Messages", "max-size")

    return server


def setup_client_object(args, config):
    ''' Reads the input arguments and the loaded configs to fill in a Client
        object containing relevant connection information

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system
    config : ConfigParser
        The set of configs found in the configuration file

    Returns
    -------
    client : rain.defaults.Client
        Contains information regarding the connection to the server
    '''
    if args.action == "set" or args.action == "get":
        interaction = "response"
        timeout_fb = DEFAULT_TIMEOUTS["Timeouts"]["receive"]
        timeout = config.getint("Timeouts", "receive", fallback=timeout_fb)

    elif args.action == "sub":
        interaction = "publish"
        timeout_fb = DEFAULT_TIMEOUTS["Timeouts"]["subscribe"]
        timeout = config.getint("Timeouts", "subscribe", fallback=timeout_fb)

    client = Client(args.server, args.action, timeout)

    client.hostname = config.get(f"{args.server}-{interaction}", "hostname")
    client.port = config.get(f"{args.server}-{interaction}", "port")

    if args.action == "sub":
        client.enable_auth = args.auth

    return client


def find_params(args):
    ''' Uses the input arguments to determine the list of parameters requested
        by the client

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system

    Returns
    -------
    params : list of strings
        The parameters requested by the client
    '''
    if args.action == "get" or args.action == "set":
        params = args.param
        logger.info(f"Parameters requested: {params}")

    elif args.action == "sub":
        params = []
        if args.changes is not None:
            for item in args.changes:
                params.append(item)
            logger.info(f"Parameter changes: {args.changes}")
        else:
            logger.info(f"Parameter changes: {None}")
        if args.freq is not None:
            for item in args.freq:
                params.append(item)
            logger.info(f"Frequency parameters: {args.freq}")
        else:
            logger.info(f"Frequency parameters: {args.freq}")
        if args.trigger is not None:
            for item in args.trigger:
                params.append(item)
            logger.info(f"Triggers: {args.trigger}")
        else:
            logger.info(f"Triggers: {None}")

        if len(params) == 0:
            logger.error("No parameters entered")
            exit()

    return params


def handle_server_args(args):
    ''' Uses the input arguments to read the server's configuration file and
        set up the necessary paths, connection details and logging

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system

    Returns
    -------
    server : rain.defaults.Server
        Contains information regarding the connection established by the server
    paths : rain.defaults.Paths
        An object containing the paths to the folders holding the senders'
        public keys, the user's private keypair and the plugins folder
    '''
    conf_folder = find_config(args)
    config = load_config(conf_folder, "server")
    setup_logging(args, config)
    paths = find_paths(config, "server")
    load_plugins(paths.plugins)
    server = setup_server_object(args, config)

    return server, paths


def handle_client_args(args):
    ''' Uses the input arguments to read the client's configuration file and
        set up the necessary paths, connection details and logging

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system

    Returns
    -------
    client : rain.defaults.Client
        Contains information regarding the connection to the server
    params : list of strings
        The parameters requested by the client
    paths : rain.defaults.Paths
        An object containing the paths to the folders holding the senders'
        public keys and the user's private keypair
    '''
    conf_folder = find_config(args)
    config = load_config(conf_folder, "client")
    setup_logging(args, config)
    paths = find_paths(config, "client")
    client = setup_client_object(args, config)
    params = find_params(args)

    return client, params, paths


def get_datetime():
    ''' Finds the current time and date with a timezone offset and returns a
        string in the ISO 8601 format

    Returns
    -------
    current_datetime : string
        The current time and date in ISO format including a timezone offset
    '''
    current_datetime = datetime.now().astimezone().isoformat()

    return current_datetime


def sub_params():
    ''' Returns a list of the parameters that the server has made available for
        clients to subscribe to

    Returns
    -------
    list_params : list of strings
        A list of parameters that can be subscribed to
    '''
    list_params = list(PLUGINS["sub"].keys())

    return list_params


def sub_trig_params():
    ''' Returns a list of the triggered parameters that the server has made
        available for clients to subscribe to

    Returns
    -------
    list_params : list of strings
        A list of trigger parameters that can be subscribed to
    '''
    list_params = list(PLUGINS["sub-trigger"].keys())

    return list_params


def get_keys(path_pub):
    ''' Returns a dict containing the public keys of all users that can connect
        and the associated user names, taken from the name of the file the key
        is stored in

    Parameters
    ----------
    path_pub : Posix path
        The path to the folder containing the public keys of the known hosts

    Returns
    -------
    keys_dict : dict
        A dict containing all server/client public keys as keys and the file
        names they are in as values
    '''
    keys_dict = {}
    for file in path_pub.glob("*"):
        key = zmq.auth.load_certificate(file)
        keys_dict[key[0].decode()] = file.stem

    return keys_dict



import json
import logging
import queue
import threading
import time

import jsonschema
import zmq

from .authenticate import setup_server
from .defaults import (
    SERVER_EXIT_KEY,
    SERVER_EXIT_CODE,
    SERVER_TRIGGER_REQ_OK,
    SERVER_TRIGGER_REQ_FAIL
)
from .fetch import (
    get_datetime,
    handle_server_args,
    sub_params,
    sub_trig_params,
)
from .packaging import (
    fill_sender_details,
    form_response,
    form_failed,
    publish_format,
    publish_update
)
from .plugins import PLUGINS
from .transport import receive_request, send_response
from .validate import validate_reqrep, validate_pub

logger = logging.getLogger(__name__)


def run_response(server, paths, exit_handler=None, exit_handler_check=10):
    ''' The function used to run all functions relevant to the handling of a
        client requesting parameters provided by this server

    Parameters
    ----------
    server : rain.defaults.Server
        Contains information regarding the connection established by the server
    paths : rain.defaults.Paths
        An object containing the paths to the folders holding the senders'
        public keys, the user's private keypair and the plugins folder
    exit_handler : function, default=None
        A function that returns a boolean to check weather the server should
        exit or not, if this is set, the socket receive will no longer be
        blocking and check the value returned according to `exit_handler_check`
    exit_handler_check : int, default=10
        Milliseconds time interval between to checks for new received messages
        and the return value of the `exit_handler` function if `exit_handler`
        is set
    '''
    socket, auth = setup_server(server, paths)

    server_open = True
    blocking = True if exit_handler is None else False
    while server_open:
        if exit_handler is not None:
            server_open = exit_handler()

        try:
            request = receive_request(socket, auth, blocking=blocking)
            logger.debug("Request received")
        except KeyboardInterrupt:
            server_open = False
            logger.info("Closing server")
            continue
        except zmq.error.Again:
            time.sleep(exit_handler_check*1e-3)
            continue

        request = fill_sender_details(auth.keys_dict, request)

        try:
            validate_reqrep(request)
        except jsonschema.exceptions.ValidationError:
            logger.error("Request validation failed:\n" + json.dumps(request))
            response = form_failed("request")
        else:
            logger.debug("Request validated")
            logger.info(f"Request: {json.dumps(request)}")
            response = form_response(request)
        finally:
            logger.debug("Response formed")

        try:
            validate_reqrep(response)
        except jsonschema.exceptions.ValidationError:
            logger.error("Response validation failed")
            response = form_failed("response")
        else:
            logger.debug("Response validated")
            logger.debug(f"Response: {json.dumps(response)}")
        finally:
            send_response(socket, response)
            logger.debug("Response sent to the client")

    auth.stop()


def run_publish(server, paths, custom_message_queue=None):
    ''' The function used to run all functions relevant to the handling of a
        client requesting parameters provided by this server

    Parameters
    ----------
    server : rain.defaults.Server
        Contains information regarding the connection established by the server
    paths : rain.defaults.Paths
        An object containing the paths to the folders holding the senders'
        public keys, the user's private keypair and the plugins folder
    custom_message_queue : queue.Queue
        Queue instance that will be used to get messages to be published by the
        server, control over this queue allows for custom injection of new
        published values and possibility to terminate the server via the magic
        `(SERVER_EXIT_KEY, SERVER_EXIT_CODE)` put into the queue
    '''
    socket, auth = setup_server(server, paths)
    possible_sub = sub_params()
    possible_sub_trig = sub_trig_params()
    if custom_message_queue is None:
        q = queue.Queue()
    else:
        q = custom_message_queue
    server_open = True

    def trigger_wait():
        ''' The worker function used by the thread running the trigger server
            to wait for a trigger parameter to be triggered
        '''
        context = zmq.Context()
        context.setsockopt(zmq.SocketOption.RCVTIMEO, 1000)
        context.setsockopt(zmq.LINGER, 0)
        tsocket = context.socket(zmq.REP)
        tsocket.bind(f"tcp://{server.trig_host}:{server.trig_port}")
        logger.debug("Trigger server opened")
        while server_open:
            try:
                trigger = tsocket.recv_json(0)
            except zmq.error.Again:
                continue
            logger.debug(f"Trigger received: {json.dumps(trigger)}")
            response = {"name": trigger["name"]}
            if trigger["name"] in possible_sub_trig:
                q.put([trigger["name"], trigger["data"]])
                response["data"] = SERVER_TRIGGER_REQ_OK
            else:
                response["data"] = SERVER_TRIGGER_REQ_FAIL
            logger.debug(f"Trigger response formed: {json.dumps(response)}")
            tsocket.send_json(response, 0)
            logger.debug("Trigger response sent to the trigger server")

    def worker(name):
        ''' The worker function run by each parameter thread, calling the
            relevant plugin function, adding the parameter value to the queue
            and waiting for the set interval before repeating this process
        '''
        func = PLUGINS["sub"][name]["function"]
        interval = PLUGINS["sub"][name]["interval"]
        while server_open:
            value = func(name)
            logger.debug(f"Sending value of '{name}'")
            q.put([name, value])
            time.sleep(interval)

    trig = threading.Thread(target=trigger_wait)
    trig.start()

    sub_threads = []
    for param in possible_sub:
        t = threading.Thread(target=worker, args=[param])
        t.start()
        sub_threads.append(t)
        logger.debug(f"Thread started for parameter {param}")

    while server_open:
        try:
            name, new_value = q.get()
        except KeyboardInterrupt:
            server_open = False
            logger.info("Stopping publish server")
            break

        if name == SERVER_EXIT_KEY and new_value == SERVER_EXIT_CODE:
            server_open = False
            logger.info("Stopping publish server from magic Queue code")
            break

        date_time = get_datetime()
        update = publish_update(name, new_value, date_time)
        logger.debug("Update formed")
        try:
            validate_pub(update)
        except jsonschema.exceptions.ValidationError:
            logger.error("Update validation failed")
        else:
            logger.debug("Update validated")
            logger.debug(f"Update: {json.dumps(update)}")
            publish = publish_format(update)
            socket.send_string(publish)
            logger.debug("Update published")

    trig.join()
    for t in sub_threads:
        t.join()

    if server.enable_auth:
        auth.stop()


def run_server(args):
    ''' The top-level function handling the function of the RAIN server

    Parameters
    ----------
    args : Namespace
        The command line arguments entered by the user
    '''
    server, paths = handle_server_args(args)

    if args.host == "rep":
        run_response(server, paths)
    elif args.host == "pub":
        run_publish(server, paths)



import logging

import zmq.auth
from zmq.auth.thread import ThreadAuthenticator

from .fetch import get_keys

logger = logging.getLogger(__name__)


class ClientCustomAuth:

    def __init__(self):
        self.key = None

    def callback(self, domain, key: bytes):
        self.key = key.decode("utf8")
        return True


def setup_auth(context, allowed, path_pub):
    ''' Set up the thread used as part of the authentication process

    Parameters
    ----------
    context : zmq.Context
        The ZMQ Context establishing the basis of the communication
    allowed : list of strings
        The hostnames of the clients that are allowed to connect to this server
    path_pub : Posix path
        The path to the folder containing the authorised public keys

    Returns
    -------
    auth : thread
        The thread that the authenticator uses to authenticate connections
    '''
    auth = ThreadAuthenticator(context)
    auth.start()
    for item in allowed:
        auth.allow(item)
    auth.configure_curve(domain="*", location=path_pub)
    auth.client_auth = ClientCustomAuth()
    auth.configure_curve_callback(domain="*", credentials_provider=auth.client_auth)
    auth.keys_dict = get_keys(path_pub)

    return auth


def setup_socket(context, host_type):
    ''' Sets the socket type depending on the kind of connection to create

    Parameters
    ----------
    context : zmq.Context
        The ZMQ Context establishing the basis of the communication
    host_type : string
        The type of connection to create: pub, req, rep, sub

    Returns
    -------
    socket : zmq.Socket
        The socket for this desired connection
    '''
    if host_type == "pub":
        socket = context.socket(zmq.PUB)
    elif host_type == "req":
        socket = context.socket(zmq.REQ)
    elif host_type == "rep":
        socket = context.socket(zmq.REP)
    elif host_type == "sub":
        socket = context.socket(zmq.SUB)

    return socket


def auth_server(socket, path_prv, auth):
    ''' Sets up the authentication side to the server connection

    Parameters
    ----------
    socket : zmq.Socket
        The connection socket
    path_prv : Posix path
        The path to the folder containing the server's private key
    auth : thread
        The thread that the authenticator uses to authenticate connections
    '''
    try:
        server_file_prv = list(path_prv.glob("*-curve.key_secret"))[0]
    except IndexError:
        logger.error("No private key file in the right format in the "
                     "keypairs folder")
        exit()

    server_pub, server_prv = zmq.auth.load_certificate(server_file_prv)
    socket.curve_secretkey = server_prv
    socket.curve_publickey = server_pub
    socket.curve_server = True
    auth.server_public_key = server_pub.decode("utf8")
    logger.debug("Server keypair loaded")


def auth_client(socket, server_name, path_pub, path_prv):
    ''' Sets up the authentication side to the client connection

    Parameters
    ----------
    socket : zmq.Socket
        The connection socket
    server_name : string
        The name of the server
    path_pub : Posix path
        The path to the folder containg the public keys of the known hosts
    path_prv : Posix path
        The path to the folder containg the client's private key

    Returns
    -------
    auth : dict
        Dict with auth information about client and server connection
    '''
    try:
        client_file_prv = list(path_prv.glob("*-curve.key_secret"))[0]
    except IndexError:
        logger.error("No private key file in the right format in the "
                     "keypairs folder")
        exit()

    client_pub, client_prv = zmq.auth.load_certificate(client_file_prv)
    socket.curve_secretkey = client_prv
    socket.curve_publickey = client_pub
    logger.debug("Client keypair loaded")

    server_file_pub = path_pub.joinpath(f"{server_name}.key")
    server_pub, _ = zmq.auth.load_certificate(server_file_pub)
    socket.curve_serverkey = server_pub
    logger.debug("Server public key loaded")

    auth = {server_pub.decode("utf8"): server_name}

    return auth


def open_connection(socket, hostname, port):
    ''' Opens the socket connection on the server side

    Parameters
    ----------
    socket : zmq.Socket
        The connection socket
    hostname : string
        The server's hostname (IP address)
    port : string
        The server's port
    '''
    socket.bind(f"tcp://{hostname}:{port}")
    logger.info(f"I am a WIP server open on {hostname} with port " +
                f"{port} ready to talk to friends")
    logger.debug("Server connection opened")


def setup_server(server, paths):
    ''' The top-level function that organises the initialisation of the server
        connection

    Parameters
    ----------
    server : rain.defaults.Server
        Contains information regarding the connection established by the server
    paths : rain.defaults.Paths
        An object containing the paths to the folders holding the senders'
        public keys, the user's private keypair and the plugins folder

    Returns
    -------
    socket : zmq.Socket
        The connection socket
    auth : thread
        The thread that the authenticator uses to authenticate connections
    '''
    context = zmq.Context()
    context.setsockopt(zmq.SocketOption.MAXMSGSIZE, server.max_msg_size)
    socket = setup_socket(context, server.host)

    if server.enable_auth:
        auth = setup_auth(context, server.allowed, paths.public)
        auth_server(socket, paths.private, auth)
    else:
        auth = None

    open_connection(socket, server.publ_host, server.publ_port)

    return socket, auth


def setup_client(client, paths):
    ''' The top-level function that organises the initialisation of the client
        connection

    Parameters
    ----------
    client : rain.defaults.Client
        Contains information regarding the connection to the server
    paths : rain.defaults.Paths
        An object containing the paths to the folders holding the senders'
        public keys and the user's private keypair

    Returns
    -------
    socket : zmq.Socket
        The connection socket
    auth : dict
        Dict with auth information about the client and server connection
    '''
    context = zmq.Context()
    context.setsockopt(zmq.SocketOption.RCVTIMEO, client.timeout)
    context.setsockopt(zmq.LINGER, 0)

    if client.action == "get" or client.action == "set":
        socket = setup_socket(context, "req")
    elif client.action == "sub":
        socket = setup_socket(context, client.action)

    if client.enable_auth:
        auth = auth_client(socket, client.server, paths.public, paths.private)
    else:
        auth = None

    return socket, auth



from .generate import gen_paths, gen_keys, gen_server_cfg, gen_client_cfg


def rain_register(args):
    ''' The top-level function handling the registration of new instruments to
        RAIN

    Parameters
    ----------
    args : Namespace
        The command line arguments entered by the user
    '''
    path_conf, path_auth, path_host, path_pair, path_plug = gen_paths(args.cfgpath, args.keypath)
    gen_keys(args.name, path_conf, path_pair)

    gen_server_cfg(path_conf, path_auth, path_pair, path_plug)
    gen_client_cfg(path_conf, path_host, path_pair)



from dataclasses import dataclass
from pathlib import Path

HOME = Path(Path.home())

# Default paths
DEFAULT_FOLDER = HOME / ".config" / "rain"
PLUGINS_FOLDER = DEFAULT_FOLDER / "plugins"
AUTHORISED_KEYS_FOLDER = DEFAULT_FOLDER / "authorised_keys"
KNOWN_HOSTS_FOLDER = DEFAULT_FOLDER / "known_hosts"
KEYPAIRS_FOLDER = DEFAULT_FOLDER / "keypairs"


DEFAULT_SERVER_CFG = {
    "Response": {
        "hostname": "127.0.0.1",
        "port": "1234"
    },
    "Publish": {
        "hostname": "127.0.0.1",
        "port": "2468"
    },
    "Trigger": {
        "hostname": "127.0.0.1",
        "port": "1793"
    },
    "Allowed": {}
}

DEFAULT_LOGGING = {
    "Logging": {
        "level": "DEBUG",
        "filepath": "None",
        "file-level": "INFO",
        "print": "True",
        "print-level": "INFO",
    }
}

DEFAULT_TIMEOUTS = {
    "Timeouts": {
        "receive": "10000",
        "subscribe": "-1"
    }
}

_CFG_PATHS_SERVER = [
    ("Security", "public-keys"),
    ("Security", "private-keys"),
    ("Plugins", "plugins")
]

_CFG_PATHS_CLIENT = [
    ("Security", "public-keys"),
    ("Security", "private-keys")
]

SERVER_EXIT_KEY = "__SERVER_EXIT__"
SERVER_EXIT_CODE = "Ezk1wDZ4MTQNG22ASim4"
SERVER_TRIGGER_REQ_OK = "Trigger received"
SERVER_TRIGGER_REQ_FAIL = "No such triggered parameter exists"

REQ_VALIDATION_ERROR = "Request verification failed"
REP_VALIDATION_ERROR = "Response verification failed"
NO_SUCH_PARAM_ERROR = "Parameter '{0}' invalid"

MAX_MESSAGE_SIZE = int(1e6)


@dataclass
class Server():
    host: str
    allowed: list
    publ_host: str = ""
    publ_port: str = ""
    trig_host: str = ""
    trig_port: str = ""
    max_msg_size: int = MAX_MESSAGE_SIZE
    enable_auth = True


@dataclass
class Client():
    server: str
    action: str
    timeout: int
    hostname: str = ""
    port: str = ""
    enable_auth = True
    pub_key: dict = None


@dataclass
class Paths():
    public: str
    private: str
    plugins: str



import json
import logging

import jsonschema
import zmq

from .authenticate import setup_client
from .fetch import handle_client_args
from .packaging import fill_sender_details, form_request, pub_split
from .transport import send_request, receive_response, receive_subscribe
from .validate import validate_reqrep, validate_pub

logger = logging.getLogger(__name__)


def run_request(client, params, data, paths):
    ''' The function used to run all functions relevant to the handling of the
        user requesting parameters provided by the server

    Parameters
    ----------
    client : Client object
        Contains information regarding the connection to the server
    params : list of strings
        The parameters to interact with
    data : list of strings
        The data to transmit along with the parameters
    paths : Path object
        An object containing the paths to the folders holding the senders'
        public keys and the user's private keypair
    '''
    request = form_request(client.action, params, data)
    logger.debug("Request formed")
    try:
        validate_reqrep(request)
    except jsonschema.exceptions.ValidationError:
        request = None
        logger.error("Request validation failed")
        return
    else:
        logger.debug("Request validated")
        logger.debug(f"Request: {json.dumps(request)}")

    if request:
        socket, client.pub_key = setup_client(client, paths)
        send_request(socket, client.hostname, client.port, request)
        logger.debug("Request sent to the server")
        try:
            response = receive_response(socket, client.hostname, client.port, client.pub_key)
        except zmq.error.Again:
            logger.error("Server not reachable, response timed out")
            return

        response = fill_sender_details(client.pub_key, response)

        logger.info("Response received from the server")
        try:
            validate_reqrep(response)
        except jsonschema.exceptions.ValidationError:
            logger.error("Response validation failed")
            return
        else:
            logger.debug("Response validated")
            logger.debug(f"Response: {json.dumps(response)}")
            yield response


def run_subscribe(client, params, paths):
    ''' The function used to run all functions relevant to the handling of the
        user subscribing to parameters provided by the server

    Parameters
    ----------
    client : Client object
        Contains information regarding the connection to the server
    params : list of strings
        The parameters to interact with
    paths : Path object
        An object containing the paths to the folders holding the senders'
        public keys and the user's private keypair
    '''
    socket, client.pub_key = setup_client(client, paths)

    for item in params:
        socket.setsockopt_string(zmq.SUBSCRIBE, item)

    socket.connect(f"tcp://{client.hostname}:{client.port}")
    logger.info("Connection opened to the publish server")

    client_connected = True
    while client_connected:
        try:
            formatted_update = receive_subscribe(socket)
        except (KeyboardInterrupt, zmq.error.Again) as e:
            if isinstance(e, KeyboardInterrupt):
                logger.info("Closing subscribe client")
            elif isinstance(e, zmq.error.Again):
                logger.error("Server not reachable, response timed out")

            client_connected = False
            for item in params:
                socket.setsockopt_string(zmq.UNSUBSCRIBE, item)
            continue

        update = pub_split(formatted_update)

        if client.pub_key:
            for key in client.pub_key:
                update["sender-key"] = key
                update["sender-name"] = client.pub_key[key]

        logger.debug(f"Update received from the server: {json.dumps(update)}")
        try:
            validate_pub(update)
            logger.debug("Update validated")
        except jsonschema.exceptions.ValidationError:
            logger.error("Update validation failed")
        else:
            yield update


def run_client(args):
    ''' The top-level function handling the function of the RAIN client

    Parameters
    ----------
    args : Namespace
        The command line arguments entered by the user
    '''
    client, params, paths = handle_client_args(args)

    if args.action == "get" or args.action == "set":
        response = run_request(client, params, args.data, paths)
    elif args.action == "sub":
        response = run_subscribe(client, params, paths)

    return response



import logging
import zmq


logger = logging.getLogger(__name__)


def send_request(socket, hostname, port, request):
    ''' Sends a request from a client to a server

    Parameters
    ----------
    socket : zmq.Socket
        The connection socket
    hostname : string
        The server's hostname (IP address)
    port : string
        The server's port
    request : JSON
        The request to send to the server
    '''
    socket.connect(f"tcp://{hostname}:{port}")
    socket.send_json(request, 0)

    return


def receive_request(socket: zmq.Socket, auth, blocking=True):
    ''' Receives a request from a client

    Parameters
    ----------
    socket : zmq.Socket
        The conenction socket

    Returns
    -------
    request : JSON
        The request sent by the client to the server
    blocking : bool
        Determines if the socket recv should be blocking or not,
        if not blocking the recv will raise `zmq.error.Again` when
        there is no new connections
    '''
    flags = 0 if blocking else zmq.NOBLOCK
    request = socket.recv_json(flags)
    request["sender-key"] = auth.client_auth.key
    return request


def send_response(socket, response):
    ''' Sends a response from a server to a client, in response to a request

    Parameters
    ----------
    socket : zmq.Socket
        The connection socket
    response : JSON
        The response to send to the client
    '''
    socket.send_json(response, 0)


def receive_response(socket, hostname, port, auth):
    ''' Receives a response from a server, in response to a request

    Parameters
    ----------
    socket : zmq.Socket
        The connection socket
    hostname : string
        The server's hostname (IP address)
    port : string
        The server's port
    auth : dict
        Dict with auth information about client and server connection

    Returns
    -------
    response : JSON
        The responsesent by the server to the client
    '''
    response = socket.recv_json(0)
    for key in auth:
        response["sender-key"] = key
    # response["sender-key"] = auth["server_public_key"]
    socket.disconnect(f"tcp://{hostname}:{port}")

    return response


def receive_subscribe(socket):
    ''' Receives an update due to the change in the value of a parameter the
        client has subscribed to

    Parameters
    ----------
    socket : zmq.Socket
        The connection socket

    Returns
    -------
    update : string
        The update sent by the server
    '''
    update = socket.recv_string()
    return update



import argparse
import pathlib

from .register import rain_register
from .trigger import rain_trigger
from .client import run_client
from .packaging import print_response
from .server import run_server

# TODO 44: Improve the layout of the --help command


def register_cli():
    ''' The CLI used for registering an instrument to RAIN

    Returns
    -------
    args : Namespace
        The command line arguments entered by the user
    '''
    parser = argparse.ArgumentParser(
        prog="RAIN",
        description="RAIN User Registration"
    )

    parser.add_argument(
        "name",
        help="the name of the instrument"
    )

    parser.add_argument(
        "-c", "--cfgpath",
        type=pathlib.PosixPath,
        help="the path to your RAIN config folder"
    )

    parser.add_argument(
        "-k", "--keypath",
        type=pathlib.PosixPath,
        help="the path to your keypair folder"
    )

    args = parser.parse_args()

    rain_register(args)


def trigger_cli():
    ''' The CLI used server-side to trigger a trigger parameter

    Returns
    -------
    args: Namespace
        The command line arguments entered by the user
    '''
    parser = argparse.ArgumentParser(
        prog="RAIN",
        description="RAIN Trigger Interface"
    )

    parser.add_argument(
        "name",
        help="the name of the trigger to send"
    )

    parser.add_argument(
        "value",
        help="the value assigned to this trigger"
    )

    parser.add_argument(
        "-c", "--cfgpath",
        type=pathlib.PosixPath,
        help="the path to your RAIN config folder"
    )

    args = parser.parse_args()

    rain_trigger(args)


def server_cli():
    ''' The CLI to start a RAIN server

    Returns
    -------
    args: Namespace
        The command line arguments entered by the user
    '''
    parser = argparse.ArgumentParser(
        prog="RAIN",
        description="RAIN Server Interface"
    )

    parser.add_argument(
        "host",
        choices=["rep", "pub"],
        help="the type of message the server will send"
    )

    parser.add_argument(
        "-a", "--auth",
        action="store_false",
        help="disables authentication for publish servers"
    )

    parser.add_argument(
        "-c", "--cfgpath",
        type=pathlib.PosixPath,
        help="the path to your RAIN config folder"
    )

    parser.add_argument(
        "-l", "--logfile",
        default=None,
        help="the path to the logfile, including its name"
    )

    parser.add_argument(
        "-o", "--logprint",
        action="store_true",
        help="whether to print logs to the console"
    )

    parser.add_argument(
        "-v", "--loglevel",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="the main logging level (first filter)"
    )

    parser.add_argument(
        "-vl", "--filelevel",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default=None,
        help="the logging level of the logfile"
    )

    parser.add_argument(
        "-vo", "--printlevel",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default=None,
        help="the logging level of the print logging"
    )

    args = parser.parse_args()

    run_server(args)


def client_cli():
    ''' The CLI to start a RAIN client

    Returns
    -------
    args : Namespace
        The command line arguments entered by the user
    '''
    parser = argparse.ArgumentParser(
        prog="RAIN",
        description="RAIN Client Interface"
    )

    parser.add_argument(
        "-s", "--suppress",
        action="store_true",
        help="whether to suppress return message printing to stdout"
    )

    parser.add_argument(
        "-c", "--cfgpath",
        type=pathlib.PosixPath,
        help="the path to your RAIN config folder"
    )

    parser.add_argument(
        "-l", "--logfile",
        default=None,
        help="the path to the logfile, including its name"
    )

    parser.add_argument(
        "-o", "--logprint",
        action="store_true",
        help="whether to print logs to the console"
    )

    parser.add_argument(
        "-v", "--loglevel",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="the main logging level (first filter)"
    )

    parser.add_argument(
        "-vl", "--filelevel",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="the logging level of the logfile"
    )

    parser.add_argument(
        "-vo", "--printlevel",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="the logging level of the print logging"
    )

    parser.add_argument(
        "server",
        help="the name of the server to connect to"
    )

    subparsers = parser.add_subparsers(
        title="RAIN Actions",
        dest="action"
    )

    parser_get = subparsers.add_parser(
        "get",
        help="get the values of parameters"
    )
    parser_get.add_argument(
        "param",
        nargs="+",
        help="the parameter(s) to get"
    )
    parser_get.add_argument(
        "-d", "--data",
        nargs="*",
        default=None,
        help="data to request alongside a parameter"
    )

    parser_set = subparsers.add_parser(
        "set",
        help="set the values of parameters"
    )
    parser_set.add_argument(
        "param",
        nargs="+",
        help="the parameter(s) to set"
    )
    parser_set.add_argument(
        "-d", "--data",
        nargs="*",
        help="data to request alongside a parameter"
    )

    parser_sub = subparsers.add_parser(
        "sub",
        help="subscribe to the values of parameters"
    )
    parser_sub.add_argument(
        "-a", "--auth",
        action="store_false",
        help="disables authentication for publish servers"
    )
    parser_sub.add_argument(
        "-c", "--changes",
        action="append",
        help="subscribe to a change in a parameter"
    )
    parser_sub.add_argument(
        "-f", "--freq",
        action="append",
        help="subscribe to all values of a parameter"
    )
    parser_sub.add_argument(
        "-t", "--trigger",
        action="append",
        help="subscribe to a triggered parameter"
    )

    args = parser.parse_args()

    messages = run_client(args)

    prev_values = []
    if args.action == "sub":
        if args.changes is not None:
            for item in args.changes:
                prev_values.append([item, ""])

    for message in messages:
        if args.action == "sub":
            if args.changes is not None:
                if message["name"] in args.changes:
                    for item in range(len(prev_values)):
                        if prev_values[item][0] == message["name"]:
                            index = item
                    if message["data"] != prev_values[index][1]:
                        prev_values[index][1] = message["data"]
                        if not args.suppress:
                            print_response(message)
                else:
                    if not args.suppress:
                        print_response(message)
            else:
                if not args.suppress:
                    print_response(message)
        else:
            if not args.suppress:
                print_response(message)



#!/usr/bin/env python

"""
Test basic kepler functions
"""
import unittest
import pathlib
import threading as th
import queue
import time

import rain
import rain.client
import rain.defaults
import rain.server
import rain.trigger
import rain.packaging


SERVER_CONFIG_LOC = pathlib.Path(".") / "examples" / "reindeer"


class TestPubServer(unittest.TestCase):

    def test_start_stop_pub_server(self):
        self.test_server = rain.defaults.Server("pub", [])
        self.test_server.publ_host = "127.0.0.1"
        self.test_server.publ_port = 8000
        self.test_server.trig_host = "127.0.0.1"
        self.test_server.trig_port = 8001

        self.paths = rain.defaults.Paths(
            SERVER_CONFIG_LOC / "authorised_keys",
            SERVER_CONFIG_LOC / "keypairs",
            SERVER_CONFIG_LOC / "plugins"
        )

        self.queue = queue.Queue()
        self.exit_magic = (rain.server.SERVER_EXIT_KEY, rain.server.SERVER_EXIT_CODE)

        server = th.Thread(
            target=rain.server.run_publish,
            args=(
                self.test_server,
                self.paths
            ),
            kwargs={
                "custom_message_queue": self.queue,
            },
        )
        server.start()
        self.queue.put(self.exit_magic)
        server.join()


class TestSubClientTowardsServer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_server = rain.defaults.Server("pub", [])
        cls.test_server.publ_host = "127.0.0.1"
        cls.test_server.publ_port = 8000
        cls.test_server.trig_host = "127.0.0.1"
        cls.test_server.trig_port = 8001

        cls.paths = rain.defaults.Paths(
            SERVER_CONFIG_LOC / "authorised_keys",
            SERVER_CONFIG_LOC / "keypairs",
            SERVER_CONFIG_LOC / "plugins"
        )

        cls.test_client = rain.defaults.Client("reindeer", "sub", 10000)
        cls.test_client.hostname = "127.0.0.1"
        cls.test_client.port = 8000

        cls.client_paths = rain.defaults.Paths(
            SERVER_CONFIG_LOC / "known_hosts",
            SERVER_CONFIG_LOC / "keypairs",
            None
        )

        rain.load_plugins(SERVER_CONFIG_LOC / "plugins")

        cls.queue = queue.Queue()
        cls.exit_magic = (rain.server.SERVER_EXIT_KEY, rain.server.SERVER_EXIT_CODE)

        cls.server = th.Thread(
            target=rain.server.run_publish,
            args=(
                cls.test_server,
                cls.paths
            ),
            kwargs={
                "custom_message_queue": cls.queue,
            },
        )
        cls.server.start()

    @classmethod
    def tearDownClass(cls):
        cls.queue.put(cls.exit_magic)
        cls.server.join()

    def test_sub_one_timed_param(self):
        response_generator = rain.client.run_subscribe(
            client=self.test_client,
            params=["activity"],
            paths=self.client_paths
        )

        response = next(response_generator)
        assert response["action"] == "sub", response
        assert response["name"] == "activity", response
        assert response["data"] == "grazing", response

        response = next(response_generator)
        assert response["action"] == "sub", response
        assert response["name"] == "activity", response
        assert response["data"] == "grazing", response

    def test_trigger_wrong_param(self):
        trigger_response = rain.trigger.send_trigger(
            server_host=self.test_server.trig_host,
            server_port=self.test_server.trig_port,
            name="activity",
            value="DANCING",
        )
        assert trigger_response["name"] == "activity", trigger_response
        assert trigger_response["data"] == rain.server.SERVER_TRIGGER_REQ_FAIL, trigger_response

    def test_trigger(self):
        trigger_response = rain.trigger.send_trigger(
            server_host=self.test_server.trig_host,
            server_port=self.test_server.trig_port,
            name="antlers",
            value="They fell off!",
        )
        assert trigger_response["name"] == "antlers", trigger_response
        assert trigger_response["data"] == rain.server.SERVER_TRIGGER_REQ_OK, trigger_response

    def test_sub_one_trigger(self):
        response_generator = rain.client.run_subscribe(
            client=self.test_client,
            params=["antlers"],
            paths=self.client_paths
        )
        local_q = queue.Queue()

        def launch_subscribe():
            response = next(response_generator)
            local_q.put(response)

        send_trigger = th.Thread(
            target=launch_subscribe,
        )
        send_trigger.start()
        time.sleep(0.01)

        trigger_response = rain.trigger.send_trigger(
            server_host=self.test_server.trig_host,
            server_port=self.test_server.trig_port,
            name="antlers",
            value="They fell off!",
        )
        assert trigger_response["name"] == "antlers", trigger_response
        assert trigger_response["data"] == rain.defaults.SERVER_TRIGGER_REQ_OK, trigger_response

        send_trigger.join()
        response = local_q.get()

        assert response["action"] == "sub", response
        assert response["name"] == "antlers", response
        assert response["data"] == "They fell off!", response



import json

from .plugins import PLUGINS, register_response


@register_response(
    action="get",
    name="api",
    data_description="api description description"
)
def generate_api(message):
    api = {
        action: [
            {"name": key, "data_description": plug["data_description"]}
            for key, plug in PLUGINS[action].items()
        ]
        for action in PLUGINS
    }
    return json.dumps(api, indent=4)



import copy
import json
import logging
import sys

from .defaults import (
    REQ_VALIDATION_ERROR,
    REP_VALIDATION_ERROR,
    NO_SUCH_PARAM_ERROR
)
from .fetch import get_datetime
from .plugins import PLUGINS

logger = logging.getLogger(__name__)


def form_request(request_type, params, data):
    ''' Creates the request message for the client to send to a server

    Parameters
    ----------
    request_type : string
        Whether the request is in the form of a GET or a SET command
    params : list of strings
        The parameters that are the subject of the request
    data : list of strings
        The data to transmit along with the parameters

    Returns
    -------
    request : JSON
        The formatted request to be sent by the client to the server
    '''
    date_time = get_datetime()
    request = {
        "sender-name": "",  # Server will fill this in to avoid spoofing
        "sender-key": "",  # Server will fill this in to avoid spoofing
        "datetime": date_time,
        "action": request_type,
        "name": params
    }

    if request["action"] == "get":
        if data is None:
            request.update({"data": ["None"]*len(params)})
        else:
            if len(params) == len(data):
                request.update({"data": data})
            else:
                logger.error("Number of data values does not match the number of parameters")
                exit()
    elif request["action"] == "set":
        if data is None:
            logger.error("Number of data values does not match the number of parameters")
            exit()
        if len(params) == len(data):
            request.update({"data": data})
        else:
            logger.error("Number of data values does not match the number of parameters")
            exit()

    return request


def form_response(request):
    ''' Creates the response message following from the reception of a request
        from a client. This response message will later be sent to the client

    Parameters
    ----------
    request : JSON
        The request made by the client

    Returns
    -------
    response : JSON
        The formatted response to be sent by the server to the client
    '''
    date_time = get_datetime()
    response = {
        "sender-name": "",  # Client will fill this in to avoid spoofing
        "sender-key": "",  # Client will fill this in to avoid spoofing
        "datetime": date_time,
        "action": request["action"],
        "name": request["name"]
    }

    data = []
    for ind, param in enumerate(request["name"]):
        try:
            func = PLUGINS[request["action"]][param]["function"]
        except KeyError:
            data.append(NO_SUCH_PARAM_ERROR.format(param))
        else:
            request_copy = copy.deepcopy(request)
            request_copy["name"] = param
            request_copy["data"] = request["data"][ind]
            try:
                return_data = func(request_copy)
            except BaseException as e:
                return_data = f"Plugin failed with: {e}"
                logger.exception("Plugin failed")
            data.append(return_data)
    response.update({"data": data})

    return response


def form_failed(form):
    ''' Creates a failure response message, either because the server received
        a request that failed validation or because its response failed
        validation. This message needs to be sent as a REQ/REP socket requires
        both a request and a response to be transmitted, otherwise blocking
        will occur

    Parameters
    ----------
    form : string
        Whether the message that failed verification was a request or a
        response

    Returns
    -------
    response : JSON
        The formatted response to be sent by the server to the client
    '''
    date_time = get_datetime()
    response = {
        "sender-name": "",  # Client will fill this in to avoid spoofing
        "sender-key": "",  # Client will fill this in to avoid spoofing
        "datetime": date_time,
        "action": "fail",
        "name": ["fail"]
    }

    if form == "request":
        response.update({"data": [REQ_VALIDATION_ERROR]})
    elif form == "response":
        response.update({"data": [REP_VALIDATION_ERROR]})

    return response


def publish_update(param, value, current_datetime):
    ''' Creates a JSON blob containing the updated value of a parameter

    Parameters
    ----------
    param : string
        The parameter whose value has been updated
    value : string
        The new values this parameter has
    current_datetime : list of strings
        The server's current local date and time

    Returns
    -------
    update : JSON
        The update to be published by the server
    '''
    update = {
        "sender-name": "",  # Client will fill this in to avoid spoofing
        "sender-key": "",  # Client will fill this in to avoid spoofing
        "datetime": current_datetime,
        "action": "sub",
        "name": param,
        "data": value
    }

    return update


def publish_format(update):
    ''' Converts the JSON update into a string and adds the parameter name and
        an extra character as a prefix, in order for the update to be detected
        by the client's filter

    Parameters
    ----------
    update : JSON
        The update to be published by the server

    Returns
    -------
    publish : string
        The reformatted update to be published by the server
    '''
    publish = f'{update["name"]}${json.dumps(update)}'

    return publish


def pub_split(publish):
    ''' Removes the prefix to the update published by the server, in order to
        recover only the JSON data

    Parameters:
    -----------
    publish : string
        The formatted update published by the server

    Returns
    -------
    update : JSON
        The JSON data containing the updated parameter
    '''
    [_, update] = publish.split('$', maxsplit=1)
    update = json.loads(update)

    return update


def print_response(response):
    ''' Prints a message sent by the server to the client in a nicer format

    Parameters
    ----------
    response : JSON
        The message sent by the server to the client
    '''
    print(json.dumps(response, indent=4, sort_keys=False))
    sys.stdout.flush()


def fill_sender_details(keys_dict, message):
    ''' Updates the received message with the sender's name and public key,
        this is done automatically on reception to counter potential spoofing

    Parameters
    ----------
    keys_dict : dict
        A dict containing sender public keys as values and their names as
        values
    message : JSON
        The message received by the receiver

    Returns
    -------
    message : JSON
        The message received by the receiver, updated to display the sender's
        name and public key
    '''
    for item in keys_dict:
        if item == message["sender-key"]:
            message["sender-name"] = keys_dict[item]

    return message



import importlib
import pathlib
import sys

PLUGINS = {
    "get": {},
    "set": {},
    "sub": {},
    "sub-trigger": {}
}


def register_response(action, name, data_description):
    ''' The function used to decorate the plugin functions relating to
        parameters that can be requested, allowing these to be registered

    Parameters
    ----------
    action : string
        The type of message the plugin function relates to, in this case "sub"
    name : string
        The name of the trigger parameter
    data_description : string
        The description of the parameter: what it is, what it does, units,
        data type, possible values, etc

    Returns
    -------
    register_wrapper : Python function
        The wrapper for registering the plugin function
    '''
    def register_wrapper(func):
        add_response(action, name, func, data_description)
        return func

    return register_wrapper


def add_response(action, name, func, data_description):
    ''' Adds a requestable parameter to the list of parameters made available
        by a server

    Parameters
    ----------
    action : string
        The type of message the plugin function relates to, in this case "sub"
    name : string
        The name of the trigger parameter
    func : Python function
        The Python plugin function
    data_description : string
        The description of the parameter: what it is, what it does, units,
        data type, possible values, etc
    '''
    global PLUGINS
    if name not in PLUGINS[action]:
        PLUGINS[action][name] = {}
    PLUGINS[action][name]["function"] = func
    PLUGINS[action][name]["data_description"] = data_description


def register_publish(action, name, interval, data_description):
    ''' The function used to decorate the plugin functions relating to timed
        parameters, allowing for these to be registered

    Parameters
    ----------
    action : string
        The type of message the plugin function relates to, in this case "sub"
    name : string
        The name of the trigger parameter
    interval : int or float
        The time interval between updates of a parameter
    data_description : string
        The description of the parameter: what it is, what it does, units,
        data type, possible values, etc

    Returns
    -------
    register_wrapper : Python function
        The wrapper for registering the plugin function
    '''
    def register_wrapper(func):
        add_publish(action, name, func, interval, data_description)
        return func

    return register_wrapper


def add_publish(action, name, func, interval, data_description):
    ''' Adds a timed parameter to the list of parameters made available by a
        server

    Parameters
    ----------
    action : string
        The type of message the plugin function relates to, in this case "sub"
    name : string
        The name of the trigger parameter
    func : Python function
        The Python plugin function
    interval : int or float
        The time interval between updates of a parameter
    data_description : string
        The description of the parameter: what it is, what it does, units,
        data type, possible values, etc
    '''
    global PLUGINS
    if name not in PLUGINS[action]:
        PLUGINS[action][name] = {}
    PLUGINS[action][name]["function"] = func
    PLUGINS[action][name]["data_description"] = data_description
    PLUGINS[action][name]["interval"] = interval


def register_trigger(name, data_description):
    ''' The function used to decorate the plugin functions relating to trigger
        parameters, allowing for these to be registered

    Parameters
    ----------
    name : string
        The name of the trigger parameter
    data_description : string
        The description of the parameter: what it is, what it does, units,
        data type, possible values, etc
    '''
    add_trigger(name, data_description)


def add_trigger(name, data_description):
    ''' Adds a trigger parameter to the list of parameters made available by a
        server

    Parameters
    ----------
    name : string
        The name of the trigger parameter
    data_description : string
        The description of the parameter: what it is, what it does, units,
        data type, possible values, etc
    '''
    global PLUGINS
    if name not in PLUGINS["sub-trigger"]:
        PLUGINS["sub-trigger"][name] = {}
    PLUGINS["sub-trigger"][name]["data_description"] = data_description


def load_plugins(plugins_folder):
    ''' Loads the files in the plugins folder, making available the functions
        they contain

    Parameters
    ----------
    plugins_folder : Posix path
        The path to the folder containg the plugin files
    '''
    if not isinstance(plugins_folder, pathlib.Path):
        plugins_folder = pathlib.Path(plugins_folder)

    if plugins_folder.exists():
        prev_sys_path = sys.path.copy()
        sys.path = [str(plugins_folder)] + sys.path
        for item in plugins_folder.iterdir():
            if item.is_file() and item.suffix == '.py':
                importlib.import_module(item.stem)
            elif item.is_dir() and (item / '__init__.py').is_file():
                importlib.import_module(item.stem)
        sys.path = prev_sys_path



import configparser
import json
from pathlib import Path

import zmq

from .defaults import DEFAULT_FOLDER


def send_trigger(server_host, server_port, name, value):
    message = {
        "name": name,
        "data": value
    }
    context = zmq.Context()
    socket = context.socket(zmq.REQ)
    socket.connect(f"tcp://{server_host}:{server_port}")
    socket.send_json(message, 0)
    response = socket.recv_json(0)
    return response


def rain_trigger(args):
    ''' Function used to send a trigger to the server's trigger server

    Parameters
    ----------
    args : Namespace
        The arguments entered into the system
    '''
    if args.cfgpath is None:
        folder = DEFAULT_FOLDER
    else:
        folder = Path(args.cfgpath)
    path_conf = folder / "server.cfg"

    config = configparser.ConfigParser()
    config.read([path_conf])

    server_host = config.get("Trigger", "hostname")
    server_port = config.get("Trigger", "port")

    response = send_trigger(server_host, server_port, args.name, args.value)

    print(json.dumps(response, indent=4, sort_keys=False))



from jsonschema import validate


def validate_reqrep(message):
    ''' Used to validate the structure of a client request or a server response
        using a validation schema

    Parameters
    ----------
    message : JSON
        The formatted message to be validated
    '''
    validate(instance=message, schema=reqrep_schema)


def validate_pub(message):
    ''' Used to validate the structure of a message from a Publish server using
        a validation schema

    Parameters
    ----------
    message : JSON
        The formatted message to be validated
    '''
    validate(instance=message, schema=pub_schema)


reqrep_schema = {
    "type": "object",
    "description": "",
    "required": ["action", "name", "data"],
    "properties": {
        "sender-name": {
            "type": "string"
        },
        "sender-key": {
            "type": "string"
        },
        "datetime": {
            "type": "string",
            "description": "the date and time at message formation, including"
            + "timezone offset, according to the ISO 8601 standard"
        },
        "action": {
            "type": "string"
        },
        "name": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "data": {
            "type": "array",
            "items": {
                "type": "string"
            }
        }
    }
}

pub_schema = {
    "type": "object",
    "description": "",
    "required": ["action", "name", "data"],
    "properties": {
        "sender-name": {
            "type": "string"
        },
        "sender-key": {
            "type": "string"
        },
        "datetime": {
            "type": "string",
            "description": "the date and time at message formation, including"
            + "timezone offset, according to the ISO 8601 standard"
        },
        "action": {
            "type": "string"
        },
        "name": {
            "type": "string"
        },
        "data": {
            "type": "string"
        }
    }
}



#!/usr/bin/env python

"""
Test of the top-level functions for running a Request Client and a Response
server
"""
import unittest
import pathlib
import threading as th

import rain
import rain.client
import rain.defaults
import rain.server
import rain.packaging


SERVER_CONFIG_LOC = pathlib.Path(".") / "examples" / "reindeer"


class TestRepServer(unittest.TestCase):

    def test_start_stop_rep_server(self):
        self.test_server = rain.defaults.Server("rep", [])
        self.test_server.publ_host = "127.0.0.1"
        self.test_server.publ_port = 8000
        self.test_server.trig_host = None
        self.test_server.trig_port = None

        self.paths = rain.defaults.Paths(
            SERVER_CONFIG_LOC / "authorised_keys",
            SERVER_CONFIG_LOC / "keypairs",
            SERVER_CONFIG_LOC / "plugins"
        )

        self.server_running = True
        server = th.Thread(
            target=rain.server.run_response,
            args=(
                self.test_server,
                self.paths
            ),
            kwargs={
                "exit_handler": lambda: self.server_running,
            },
        )
        server.start()
        self.server_running = False
        server.join()


class TestRepClientTowardsServer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_server = rain.defaults.Server("rep", [])
        cls.test_server.publ_host = "127.0.0.1"
        cls.test_server.publ_port = 8000
        cls.test_server.trig_host = None
        cls.test_server.trig_port = None

        cls.server_paths = rain.defaults.Paths(
            SERVER_CONFIG_LOC / "authorised_keys",
            SERVER_CONFIG_LOC / "keypairs",
            SERVER_CONFIG_LOC / "plugins"
        )

        cls.test_get_client = rain.defaults.Client("reindeer", "get", 1000)
        cls.test_get_client.hostname = "127.0.0.1"
        cls.test_get_client.port = 8000

        cls.test_set_client = rain.defaults.Client("reindeer", "set", 1000)
        cls.test_set_client.hostname = "127.0.0.1"
        cls.test_set_client.port = 8000

        cls.client_paths = rain.defaults.Paths(
            SERVER_CONFIG_LOC / "known_hosts",
            SERVER_CONFIG_LOC / "keypairs",
            None
        )

        rain.load_plugins(cls.server_paths.plugins)

        cls.server_running = True
        cls.server = th.Thread(
            target=rain.server.run_response,
            args=(
                cls.test_server,
                cls.server_paths
            ),
            kwargs={
                "exit_handler": lambda: cls.server_running,
            },
        )
        cls.server.start()

    @classmethod
    def tearDownClass(cls):
        cls.server_running = False
        cls.server.join()

    def test_get_wrong_param(self):
        response_generator = rain.client.run_request(
            client=self.test_get_client,
            params=["what_they_doin"],
            data=None,
            paths=self.client_paths
        )
        response = next(response_generator)
        assert response["data"][0] == rain.defaults.NO_SUCH_PARAM_ERROR.format("what_they_doin"), response

    def test_get_one_param(self):
        response_generator = rain.client.run_request(
            client=self.test_get_client,
            params=["activity"],
            data=None,
            paths=self.client_paths
        )

        response = next(response_generator)
        assert response["action"] == "get", response
        assert response["name"][0] == "activity", response
        assert response["data"][0] == "grazing", response

    def test_get_two_params(self):
        response_generator = rain.client.run_request(
            client=self.test_get_client,
            params=["herd_size", "activity"],
            data=None,
            paths=self.client_paths
        )

        response = next(response_generator)
        assert response["action"] == "get", response
        assert response["name"][0] == "herd_size", response
        assert response["data"][0] == "10", response
        assert response["name"][1] == "activity", response
        assert response["data"][1] == "grazing", response

    def test_set_one_param(self):
        response_generator = rain.client.run_request(
            client=self.test_set_client,
            params=["activity"],
            data=["sleeping"],
            paths=self.client_paths
        )

        response = next(response_generator)
        assert response["action"] == "set", response
        assert response["name"][0] == "activity", response

        # Run a `get` command to verify the `set` command worked
        response_generator = rain.client.run_request(
            client=self.test_get_client,
            params=["activity"],
            data=None,
            paths=self.client_paths
        )

        response = next(response_generator)
        assert response["action"] == "get", response
        assert response["name"][0] == "activity", response
        assert response["data"][0] == "sleeping", response

        # Set the modified value back to its default
        response_generator = rain.client.run_request(
            client=self.test_set_client,
            params=["activity"],
            data=["grazing"],
            paths=self.client_paths
        )

    def test_set_two_params(self):
        response_generator = rain.client.run_request(
            client=self.test_set_client,
            params=["herd_size", "activity"],
            data=["14", "sleeping"],
            paths=self.client_paths
        )

        response = next(response_generator)
        assert response["action"] == "set", response
        assert response["name"][0] == "herd_size", response
        assert response["name"][1] == "activity", response

        # Run a `get` command to verify the `set` command worked
        response_generator = rain.client.run_request(
            client=self.test_get_client,
            params=["herd_size", "activity"],
            data=None,
            paths=self.client_paths
        )

        response = next(response_generator)
        assert response["action"] == "get", response
        assert response["name"][0] == "herd_size", response
        assert response["data"][0] == "14", response
        assert response["name"][1] == "activity", response
        assert response["data"][1] == "sleeping", response

        # Set the modified values back to their defaults
        response_generator = rain.client.run_request(
            client=self.test_set_client,
            params=["herd_size", "activity"],
            data=["10", "grazing"],
            paths=self.client_paths
        )
